<?php
/**
 * The database_object is an object database mapper for mysql and AdoDB (e.g.
 * Postre SQL)
 *
 * This class provides a layer to access a specified table of the database. 
 * Name of the table is specified by the name of the subclass. 
 *
 * Attributes are automatically determined from database and stored in the 
 * instance.
 *
 * @author Jonas Mitschang, jonen@mitschang.net
 * http://mitschang.net/software/dbo.html
 *
 * Changelog:
 * 1.26
 *   - Optional "reload_after_insert": Default columns will be added by the 
 *     database and thus have to be reloaded after insert.
 *   - PDO database adapter (tested for psql).
 *   - support more datatypes and checks between (e.g. mysql bit(1) is PHP bool)
 *   - attributes may be statically defined, thus "SHOW FIELDS FROM" is not 
 *     needed to be called at runtime.
 * 1.25
 *   - datatype checks on attribute assignments
 *   - AdoDB insert/update without autoexecute (because it fails in some cases)
 * 1.24
 *   - more limit options (e.g. get_all(..., 5) returns the first five objects)
 * 1.23
 *   - minor changes (cache, mysql table quotings)
 * 1.22
 *   - db_adodb can create NEW connections now, using user/pass
 *     (instead of using an existing connection)
 *   - logging functionality using ::set_logger()
 *   - fetching additional rows (e.g. SELECT user.*, RANDOM() AS foo FROM ...)
 * 1.21 
 *   - db_adodb added
 *   - mysql adapter can be initialized with an existing connection
 *   - bugfix when using prefix (e.g. joomla jos_...)
 * 1.2
 *   - first public release
 */

class database_object {
    /**
     * Automatically save changes off attributes to the database? 
     */
    public $autosave = false;

    /** 
     * Oject exists in database
     */
    public $existent = false;

    /**
     * Needed because default columns will be added by the database and thus
     * have to be reloaded after insert.
     */
    public $reload_after_insert = false;

    /**
     * Column definitions 
     * e.g. var $coldef = array('user0_id' => 'user', 'user1_id' => 'user);
     */
    protected $coldef = array();

    //! @cond
    /** 
     * Hash with all attributes, that this database object contains in the 
     * database. The hash has the following format: 
     *   name => datatype
     * Do not make this static - will not work because all database_objects 
     * would have the same attributes in this case.
     */
    protected $_attributes = null;

    /** 
     * Hash with all attributes, that this database object contains in the
     * database. The hash has the following format: 
     *   attribute name => value
     */
    protected $_values = array();

    /** Hash with all (loaded) sub-objects
     */
    private $_subobj;

    /** modified fields are marked in here */
    protected $modified;

    /** enable type checking for variable assignments? */
    private static $_enable_typechecking = true;

    /** 
     * The name of this database object and the corresponding table.
     * Do not make this static - will not work because all database_objects 
     * would have the same name in this case.
     */
    protected $table_name = null;

    /**
     * Static cache for attributes and objects 
     */
    static protected $_cache = array();

    /**
     * database connection
     */
    static private $db = null;

    /**
     * enable debugging?
     */
    static public $_debug = false;
    //! @endcond

    /**
     * Constructor
     *
     * @param $id
     *   If id given: fetch the corresponding row from database
     *     Example: $vehicle = new vehicle(10); // fetch vehicle where id=10
     *   If not: create an empty object, that may be saved lateron (INSERT)
     *     Example: $new_v = new vehicle(); $new_v->color = 'red'; $new_v->save();
     */
    function __construct($id = null) {
        // db is checked here once - so no further checking needed
        assert('isset(self::$db)'); 

        // which table to use?
        $this->table_name = self::$db->prefix.$this->get_name();

        // get the attributes list (if not predefined)
        if (is_null($this->_attributes))
            $this->_attributes = self::load_attribute_list($this->get_name());

        // only load if a valid $id is given
        if (array_key_exists("id", $this->_attributes)) $this->set_with_type('id', 0);
        if (!isset($id) || is_null($id)) return;
        assert('is_numeric($id)'); // TODO is_int later

        // check cache and load from cache if existing
        $cached = $this->get_cache($id);
        if (!is_null($cached)) {
            foreach ($cached as $k => $v) $this->$k = $v;
            return;
        } 

        // load all database attrobutes
        if ($this->load_attributes($id)) $this->existent = true;
        // TODO irgendwie speichert der das hier noch nicht - also wird trotzdem 
        // geladen beim naechsten mal. er speichert eine null an __dbo_test__1
        $this->set_cache($id, $this);
    }

    public function __destruct() {
        // foreach ($this->get_attributes() as $attr => $type) {
        //     if (is_object($this->$attr)) {
        //         $this->$attr->__destruct();
        //     }
        //     unset($this->$attr);
        // }
        // unset makes problems when serializing to session
        // unset($this->_attributes);
        // unset($this->_values);
    }

    /**
     * Is the ORM already initialized with an db_adaper?
     */
    public static function is_initialized() {
        return !is_null(self::$db);
    }

    /**
     * Initialize DBO ORM mapper using a db_adapter database connection
     */
    public static function init($db) {
        assert('$db instanceof db_adapter');
        self::$db = $db;

        // init database adapter
        self::$db->init();
    }

    /**
     * Return all attribues as hash: attribute => type
     * Where type is the mysql type (e.g. integer, bool, float, string)
     */
    public function get_attributes() {
        assert('is_array($this->_attributes)');
        return $this->_attributes;
    }

    /**
     * Get the datatype of one attribute (e.g. string / integer)
     */
    public function get_attribute_type($attr) {
        assert('is_string($attr)');

        // exact column match?
        if ($this->attribute_exists($attr)) {
            return $this->_attributes[$attr];
        }

        // match with _id ?
        if ($this->attribute_exists($attr . '_id')) {
            return 'database_object';
        }
        assert('false /* attribute does not exist: '.$attr.' */');
    }

    public function get_name() {
        return get_class($this);
    }

    /**
     * Write and read the cache
     */
    private function get_cache($var) {
        return self::_get_cache($this->get_name(), $var);
    }
    static protected function _get_cache($name, $var) {
        if (self::$_debug) echo "  get cache: ".$name." ".$var."\n";
        if (is_null(self::$_cache)) return null;
        $name = $name.'__'.$var;
        if (!array_key_exists($name, self::$_cache)) return null;
        if (self::$_debug) echo "    HIT\n";
        return self::$_cache[$name];
    }
    protected function set_cache($var, $val) {
        self::_set_cache($this->get_name(), $var, $val);
    }
    static protected function _set_cache($name, $var, $val) {
        if (is_null(self::$_cache)) return;
        if (self::$_debug) echo "  set cache: ".$name." ".$var."\n";
        assert('is_array(self::$_cache)');
        $name = $name.'__'.$var;
        self::$_cache[$name] = $val;
    }

    /**
     * Get the name (the table name) of this database object
     */
    final public function get_table_name() {
        return $this->table_name;
    }

    /**
     * Return the database adapter
     */
    public static function get_db() {
        return self::$db;
    }

    protected function add_attribute($attr) {
        $this->_attributes[$attr] = 'no_db';
    }

    // Build the SQL-Query to load all attributes from Database
    private function select_statement($where = null, $limit = null, $order = null) {
        $query = 'SELECT * FROM '.self::$db->quote_table_name($this->table_name);
        $query .= self::$db->get_where($where);
        $query .= self::$db->get_order($order);
        $query .= self::$db->get_limit($limit);
        $query .= ';';
        return $query;
    }

    // Not documented => not in doxy. See get_all
    function _get_all($where = null, $limit = null, $order = null) {
        return $this->_get_all_query($this->select_statement($where, $limit, $order));
    }

    static function select_single($query) {
        return self::$db->select_single($query);
    }
    static function select_row($query) {
        return self::$db->query_row($query);
    }

    // Not documented => not in doxy. See get_all_query
    function _get_all_query($query) {
        assert('is_string($query)');

        // create objects
        $all = self::$db->query_all($query);
        $result = array(); 
        $obj_name = $this->get_name();
        foreach ($all as $item) {
            // cached?
            //if ($this->attribute_exists('id')) {
            //    $obj = self::get_cache($obj_name, $item['id']);
            //} else {
            //    $obj = null;
            //}
            //if (is_null($obj)) {
                // not cached: create new object
                $obj = new $obj_name();
                $obj->assign_record_to_attributes($item);

                // object exists in database
                $obj->existent = true;

                // save cache
                if ($this->attribute_exists('id')) {
                    $this->set_cache($obj->id, $obj);
                }
            //}

            // save to result array
            array_push($result, $obj);
        }
        return $result;
    }

    /**
     * Check if the given attribute exists
     */
    function attribute_exists($attr) {
        // called on object
        $attributes = $this->_attributes;
        return isset($attributes[$attr]);
    }
    static function attribute_exists_static($attr) {
        // called static
        $attributes = self::load_attribute_list(get_called_class());
        return isset($attributes[$attr]);
    }

    static private function load_attribute_list($name) {
        // first, try loading from cache
        $table_name = (self::$db->prefix).$name;
        $attributes = self::_get_cache($name , '_attributes');
        if (is_null($attributes)) {
            // attributes not cached => fetch from database
            $attributes = self::$db->fetch_columns($table_name);

            // and store to cache
            self::_set_cache($name, '_attributes', $attributes);
        }
        return $attributes;
    }

    /**
     * Return the type of an attribute
     */
    function attribute_type($attr) {
        assert('isset($this->_attributes[$attr])');
        return $this->_attributes[$attr];
    }

    private function load_owners($what=null, $order='id ASC', $force=false) {
        if (!isset($what)) {
            // load all
            assert('false /*loading all owners is not implemented yet - and never will be */');
        }

        if (is_array($what)) {
            // load multiple attributes
            foreach ($what as $attr) $this->load_owners($attr);
        } else {
            // load one owner
            $owner = new $what();

            // check if we really belong to owner
            $our_id = $this->get_name().'_id';
            if (!$owner->attribute_exists($our_id)) 
                assert('false /* Cannot load owners: "'.$what.'" has no attribute "'.$our_id.'" */');

            // already loaded? nothing to do
            if (!$force && isset($this->_subobj[$what])) return;

            // load (ordered by id per default)
            if ($this->id == 0) {
                $this->_subobj[$what] = array();
            } else {
                $this->_subobj[$what] = $owner->_get_all(
                    array($this->get_name() => $this), null, $order);
            }
        }
    }

    /**
     * Loads objects from a db-table to a local field
     *
     * Example:
     *   $o->load('vehicle') uses the column vehicle_id to load the
     *   vehicle with that id from the vehicle table. The resulting
     *   vehicle object is stored to $obj->vehicle
     *   Resulting query: SELECT * FROM vehicle WHERE id=$o->vehicle_id
     */
    private function load($what) {
        if (is_array($what)) {
            // load multiple attributes
            foreach ($what as $attr) $this->load($attr);
            return;
        } 

        // load one attribute
        $what_id = $what . '_id';
        assert('$this->attribute_exists($what_id)');
        $objname = $this->get_subobj_class($what);
        // TODO class_exists ?

        // id given?
        $id = $this->_values[$what_id];
        if (!isset($id)) {
            // no id => nothing to load => unset
            $this->_subobj[$what] = new $objname();
            return;
        }

        // already loaded (the same)? noting to do
        if (isset($this->_subobj[$what]) && ($id == $this->_subobj[$what]->id)) return;

        // got id => laod from db
        // shortcut: try to fetch from cache
        $new_obj = self::_get_cache($what, $id);
        if (!isset($new_obj)) $new_obj = new $objname($id);
        $this->_subobj[$what] = $new_obj;
    }

    public function attributes_type($attr) {
        echo "attributes_type deprecated: use get_attribute_type instead\n";
        return $this->get_attribute_type($attr);
    }

    private function set_with_type($attr, $value) {
        assert('isset($this->_attributes[\''.$attr.'\'])');

        // assert integer / float etc
        $type = $this->get_attribute_type($attr);

        // bool allowes 't' / 'f' for true and false and int values 
        // but you should not use that. simply use true and false
        if ($type == 'bool') {
            if ($value === 't') {
                echo 'DEPRECATED: assigning: '.$this->get_name().'->'.$attr.'(bool) = '.$value.'('.gettype($value).') 2011-06'."\n";
                $value = true;
            } elseif ($value === 'f') {
                echo 'DEPRECATED: assigning: '.$this->get_name().'->'.$attr.'(bool) = '.$value.'('.gettype($value).') 2011-06'."\n";
                $value = false;
            } elseif (is_int($value)) {
                $value = $value ? true : false; 
            }
        }

        if (self::$_enable_typechecking) {
            switch ($type) {
            case 'bool':
                assert('is_bool($value) || is_null($value) /* assigning: '.$this->get_name().'->'.$attr.'(bool) = '.$value.'('.gettype($value).') */');
                break;
            case 'integer': 
                // TODO (int)$value == $value statt ctype_digit
                // ctype_digit does not work for e.g. "-5"
                // assert('is_int($value) || ctype_digit($value) || is_null($value) /* '.$this->get_name().'->'.$attr.'('.$type.') = '.var_export($value, true).' */');
                if (is_string($value)) $value = (int)$value;
                break;
            case 'float': 
                // assert('is_numeric($value) || is_null($value) /* assigning to "'.$attr.'" */');
                if (is_string($value)) $value = (float)$value;
                break;
            case 'string': 
                //assert('is_string($value) || is_null($value) /* assigning: '.$this->get_name().'->'.$attr.'(string) = '.$value.'('.gettype($value).') */');
                if (!is_string($value) && !is_null($value))
                    echo 'DEPRECATED: assigning: '.$this->get_name().'->'.$attr.'(string) = '.$value.'('.gettype($value).') 2011-06'."\n";
                break;
            // TODO datetime assertions ...
            }
        }
        $this->_values[$attr] = $value;
    }

    private function assign_record_to_attributes($record) {
        foreach ($record as $attr => $value) {
            if (!isset($this->_attributes[$attr])) {
                // fetched an additional database row => add attribute 
                $this->add_attribute($attr);
            } else {
                // field in database => which type has the column?
                $type = $this->get_attribute_type($attr);

                // some adaper implementations return 't' and 'f' for bools
                // this sould better be moved into the adapter implementation but 
                // the adapter does not know the type (get_attribute_type)
                if ($type == 'bool') {
                    if ($value === 't') $value = true;
                    else if ($value === 'f') $value = false;
                    else if ($value === chr(1)) $value = true;
                    else if ($value === chr(0)) $value = false;
                }
            }

            // store this attribute
            $this->set_with_type($attr, $value);
        }
    }

    //! @cond
    /**
     * Method load
     * This method loads all attributes from a db-table. The table's name will
     * be extracted from the name of the subclass.
     */
    protected function load_attributes($id) {
        // select from SQL server
        $query = $this->select_statement('id='.$id);
        $record = self::$db->query_row($query);

        // not found in database
        if (!$record) return false;

        // found
        $this->assign_record_to_attributes($record);
        assert('$this->id == $id');
        return true;
    }
    //! @endcond

    /**
     * Saves the database-object into the database.
     * If existent, then call the update-method, if not, add a new entry
     */
    public function save() {
        assert('array_key_exists("id", $this->_attributes) /* no id column, save() prohibited */');
        if ($this->existent) {
            $this->update();
        } else  {
            $this->insert();
        }

        // now the object must be existent
        assert('$this->existent');
        assert('$this->id > 0');

        // also update the cache if the object is accessed later on
        $this->set_cache($this->id, null);
        // FIXME: nicht ganzen cache loeschen nur das element, aber geht 
        // irgendwie nicht
        //self::cache_clear();
    }

    /**
     * Return an array of parameters that have modified (not comitted to 
     * database)
     */
    function get_modified() {
        $result = array();
        if (is_null($this->modified)) return $result;
        foreach ($this->modified as $attr => $mod) {
            if ($mod) array_push($result, $attr);
        }
        return $result;
    }

    /**
     * Clear the cache. Should normally not be used - just in case something 
     * behaves 'strange'
     */
    static public function cache_clear() {
        if (self::$_debug) echo "  clear cache\n";
        self::$_cache = array();
    }

    /**
     * Set the query logger
     */
    public static function set_logger($fn) {
        self::$db->set_logger($fn);
    }

    /**
     * Un-cache the sub-objects (foreign keys)
     */
    public function uncache_subobj() {
        $this->_subobj = null;
    }

    /**
     * Disable the cache.
     * Can be used if plenty of objects will be kelled and thus the php memory 
     * limit would be hit.
     */
    static public function cache_disable() {
        self::$_cache = null;
    }

    /**
     * Disable variable type checking. You should not use this.
     */
    static public function typechecking_disable() {
        self::$_enable_typechecking = false;
    }

    //! @cond
    /**
     * Method update
     * 
     * Updates the representing row in the database with current attributes of
     * the database-object
     */
    protected function update() {
        assert('$this->existent /* trying to update inexistent database object. use save() */');
        assert('$this->id > 0');

        // no changes => nothing to do
        if (!isset($this->modified)) return;
        assert('is_array($this->modified)');
        if (sizeof($this->modified) == 0) return;

        // what to update?
        $record = array();
        foreach ($this->modified as $attr => $mod) {
            if ($mod) $record[$attr] = $this->$attr; 
        }
        self::$db->update($this->table_name, $record, $this->id);

        // okay no modifications now
        $this->modified = array();
    }
    //! @endcond

    /**
     * Optimize this table. Call e.g. from Admin interface or once on long 
     * program run.
     */
    public static function optimize() {
        self::$db->optimize($this->table_name);
    }

    /**
     * Dump this object to console for debuging purposes
     */
    public function dump($depth = 0) {
        if ($depth > 2) return;
        $indent = '';
        for ($i = 0; $i < $depth * 4; $i++) $indent .= ' ';
        if ($depth == 0) {
            echo $indent . $this->get_name() . '(' . $this->id . '):';
            if (!$this->existent) echo "not existent";
            echo "\n";
        }
        foreach ($this->_attributes as $k => $t) {
            if ($k == 'id') continue;
            echo $indent . '  - ' . $k . ' ('.$t.') = ' . $this->$k . "\n";
            // TODO subobj dump
            if (0) if (substr($k, -3, 3) == '_id') {
                $obj_name = substr($k, 0, strlen($k) - 3);
                $class_name = $this->get_subobj_class($obj_name);
                if (!class_exists($class_name)) {
                    echo  $indent . '    class ' . $class_name . ' not existing'."\n";
                } else {
                    $obj = $this->$obj_name;
                    $obj->dump($depth + 1);
                }
            }
        }
    }

    /**
     * Determine which class a sub object is
     */
    private function get_subobj_class($what) {
        if (isset($this->coldef[$what])) return $this->coldef[$what];
        return $what;        
    }


    /**
     * Delete the representing row in database 
     */
    public function delete() {
        assert('$this->existent /* ERROR: trying to delete inexistent database object */');
        assert('$this->id > 0');

        self::$db->query('DELETE FROM '.$this->table_name.' WHERE id='.$this->id);

        // also remove from cache
        $this->set_cache($this->id, null);

        // okay, no longer in database
        $this->id = null;
        $this->existent = false;
    }

    //! @cond
    /**
     * Adds a new row to database with current attributes
     * of the database-object
     */
    protected function insert() {
        $record = array();
        foreach ($this->_attributes as $attr => $type) {
            $value = $this->$attr;
            if (($attr == 'id') && ($value === 0)) continue; // id=0 =>autoindex
            if (!isset($value)) continue; // not set => default
            $record[$attr] = $value;
        }

        // okay created => store id
        $this->id = self::$db->insert($this->table_name, $record);

        // okay no modifications now
        $this->modified = null;
        $this->existent = true;

        // need to reload after insert?
        if ($this->reload_after_insert) $this->load_attributes($this->id);
    }
    //! @endcond

    // Magic methods
    public function __get($k) {
        if ($this->attribute_exists($k)) {
            // return the field's value
            if (!isset($this->_values[$k])) return null;
            return $this->_values[$k];
        } else if (isset($this->_subobj[$k])) {
            return $this->_subobj[$k];
        }

        // load sub object attribute
        $k_id = $k . "_id";
        if ($this->attribute_exists($k_id)) {
            $this->load($k);
            return $this->_subobj[$k];
        } 

        // not found: has_many?
        if (isset($this->_subobj[$k])) return $this->_subobj[$k];
        if (isset($this->has_many)) $this->load_owners($this->has_many);
        if (isset($this->_subobj[$k])) return $this->_subobj[$k];

        // not found
        assert('false /* Property "'.$k.'" not found in the database_object "'.
            $this->table_name.'" (reading). Add column "'.$k.'" to table "'.
            $this->table_name.'" */');
    }
    public function __set($k, $v) {
        if ($this->attribute_exists($k)) {
            // update the modified field
            if (!array_key_exists($k, $this->_values) || ($this->_values[$k] !== $v)) {
                $this->modified[$k] = true; 

                // update the field itself and save if desired
                $this->set_with_type($k, $v);
                //$this->_values[$k] = $v;
                if ($this->autosave && $this->existent) $this->save();
            }

            // return
            return $this->_values[$k];
        } 

        // load one attribute
        $k_id = $k . "_id";
        if ($this->attribute_exists($k_id)) {
            if ($v === null) {
                // null object is stored as _id = null
                $this->$k_id = null;
            } else {
                // update _id field
                assert('$v instanceof database_object');
                $this->$k_id = $v->id;
            }

            // object was loaded? reload
            if (isset($this->_subobj[$k])) {
                $this->load($k);
            }
            return $this->_subobj[$k];
        } 

        // not found
        assert('false /* Property "'.$k.'" not found in the database_object "'.
            $this->table_name.'" (writing). Add column "'.$k.'" to table "'.
            $this->table_name.'" */');
    }
    public function __isset($k) {
        return isset($this->_values[$k]);
    }
    public function __unset($k) {
        unset($this->_values[$k]);
    }

    /**
     * Is the given attribute changed (not yet saved to the database)
     */
    public function is_modified($attr) {
        assert('is_string($attr)');
        if ($this->modified == null) return false;
        if (!isset($this->modified[$attr])) return false;
        return $this->modified[$attr] != false;
    }

    public function __toString() {
        if ($this->existent) {
            return $this->get_name() . $this->id;
        }
        return $this->get_name() . '?';
    }

    /**
     * Load a bunch of database objects from the corresponding table.
     *
     * @param $where
     *   The SQL 'WHERE' part of the select statement. Set to null or 'TRUE' to 
     *   select all objects. Use array notation for better readabilty.
     *   Example: Get all red vehicles that use diesel:
     *     vehicle::get_all(array(
     *       'color' => 'red', 
     *       'fuel' => 'diesel'
     *     ));
     *
     * @param $limit
     *   The SQL 'LIMIT' part of the select statement. Set to null to get all 
     *   objects with no limit. 
     *   The first row is identified (like normal for SQL) with 'start'=>0.
     *
     * @param $order
     *   The SQL 'ORDER' part of the select statement. Set to null to irgnore
     *   ordering.
     *
     * @return an array of databse objects foo. The array may be empty if there 
     *   are no matching objects.
     *
     * Example: 
     *   Get first ten employees, ordered by login name:
     *   $all = employee::get_all(
     *     null, array('start'=>0, 'count'=>10), 'login ASC');
     */
    public static function get_all($where = null, $limit=null, $order=null) {
        // has to be called from static context
        $obj_name = get_called_class();
        $obj = new $obj_name();
        return $obj->_get_all($where, $limit, $order);
    }

    /**
     * Return the last-created object (this is by definition the object with the
     * highest id value)
     *
     * @param $where
     *   If where is set, then the last object matching the specified WHERE 
     *   condition will be returned.
     *
     * @returns The last database-object ih this table. If there is no matching 
     *   object, the function will return null.
     */
    public static function get_last($where = null) {
        return self::get_single($where, array('start'=>0, 'count'=>1), "id DESC");
    }

    /**
     * Return an array of objects that result from the given query
     * Accordingly SELECTs over multiple tables can be used
     */
    public static function get_by_query($query, $limit = null) {
        assert('is_string($query)');
        $obj_name = get_called_class();
        $obj = new $obj_name();
        $query .= self::$db->get_limit($limit);
        return $obj->_get_all_query($query, $limit);
    }

    /**
     * Return the only object that matches the expression
     * The code checks, that there is only one match.
     * If there is no match, null is returned.
     */
    public static function get_single($where=null, $limit=null, $order=null) {
        $all = self::get_all($where, $limit, $order);
        assert('is_array($all)');
        assert('sizeof($all) <= 1 /* size='.sizeof($all).' */');
        if (sizeof($all) == 0) return null;
        return $all[0];
    }

    /**
     * Count the number of objects that match the given condition.
     * @see get_all to learn more abit the conditions ($where)
     *
     * @param where Condition like described in get_all
     * @returns The number of matches as integer
     */
    public static function count($where = null) {
        // has to be called from static context
        $obj_name = get_called_class();
        $query = 'SELECT COUNT(*) AS count FROM '.(self::$db->prefix).$obj_name.
            self::$db->get_where($where).';';
        return (int)self::select_single($query);
    }

    /**
     * Count number of rows that would be returned by the given query
     */
    public static function count_query($query = null) {
        // TOOD assert: ohne LIMIT
        //$query = preg_replace('/SELECT[ -~\xA\xD][ -~\xA\xD.]*?[ -~\xA\xD]FROM/i', 'SELECT COUNT(*) FROM', $query, 1);
        //$query = preg_replace('/ORDER BY.*/', '', $query, 1);
        $query = 'SELECT COUNT(*) FROM ('.$query.') AS tab';
        return (int)self::select_single($query);
    }

    /**
     * Get the corresponding object by its id.
     * Id may be null or 0 - accordingly null will be returned
     * Thats the difference to the new opeartor.
     */
    public static function get_by_id($id) {
        assert('is_int($id) || is_null($id)');
        if (is_null($id) || $id == 0) return null;
        $obj_name = get_called_class();
        $result = new $obj_name($id);
        assert('$result->existent /* ' . $obj_name . '.id == ' . $id . ' */');
        return $result;
    }
}

/**
 * Retro-support of get_called_class()
 * (PHP 5 >= 5.3.0)
 * get_called_class — the "Late Static Binding" class name
 */
if (!function_exists('get_called_class')) {
    function get_called_class($bt = false,$l = 1) {
        if (!$bt) $bt = debug_backtrace();
        if (!isset($bt[$l])) {
            throw new Exception(
                "Cannot find called class -> stack level too deep.");
        }
        if (!isset($bt[$l]['type'])) {
            throw new Exception ('type not set');
        }
        switch ($bt[$l]['type']) {
        case '::':
            $lines = file($bt[$l]['file']);
            $i = 0;
            $callerLine = '';
            do {
                $i++; 
                $callerLine = $lines[$bt[$l]['line']-$i] . $callerLine;
            } while (stripos($callerLine,$bt[$l]['function']) === false);
            preg_match('/([a-zA-Z0-9\_]+)::'.$bt[$l]['function'].'/',
                $callerLine,
                $matches);
            if (!isset($matches[1])) {
                // must be an edge case.
                throw new Exception ("Could not find caller class: originating method call is obscured.");
            }
            switch ($matches[1]) {
            case 'self':
            case 'parent':
                return get_called_class($bt,$l+1);
            default:
                return $matches[1];
            }
            // won't get here.
        case '->': switch ($bt[$l]['function']) {
            case '__get':
                // edge case -> get class of calling object
                if (!is_object($bt[$l]['object'])) throw new Exception ("Edge case fail. __get called on non object.");
                return get_class($bt[$l]['object']);
            default: return $bt[$l]['class'];
        }

        default: throw new Exception ("Unknown backtrace method type");
        }
    }
}

