<?php
abstract class db_adapter {
    var $prefix;
    protected $table_delimiter = '`';
    protected $query_logger_callback = null;
    protected $query_count = 0;

    /**
     * Set the prefix that should be used.
     * e.g. joomla uses only tables starting with 'jos_', thus 
     * set_prefix('jos_') can be used for simplification.
     */
    function set_prefix($prefix) {
        $this->prefix = $prefix;
    }

    /**
     * Get total numer of SQL queries
     */
    function get_query_count() {
        return $this->query_count;
    }

    /**
     * Log and count one query
     * @param $post if false before the query and true after the query
     */
    function log_query($query, $post = false) {
        if (!is_null($this->query_logger_callback)) {
            call_user_func($this->query_logger_callback, $query, $post);
        }
        $this->query_count++;
    }

    /**
     * Quote the table name
     * Will be overwritten specifically by the adapters
     */
    function quote_table_name($table_name) {
        return $table_name;
    }

    /**
     * Set database logging function.
     * ->set_logger('foo') calls global function foo($query)
     * ->set_logger(array('foo', 'bar')) calls static function foo::bar($query)
     */
    function set_logger($fn) {
        $this->query_logger_callback = $fn;
    }

    /**
     * Get the LIMIT part of an SQL query.
     * @param $limit = null means no limit.
     *   $limit = array('start' => 0, 'count' => 10)
     */
    public function get_limit($limit) {
        if (is_null($limit)) {
            return '';
        } else if (is_int($limit)) {
            return ' LIMIT 0, '.$limit;
        } else if (is_array($limit)) {
            if (isset($limit['start'])) {
                $start = $limit['start'];
            } else {
                $start = 0;
            }
            assert('is_numeric($start)');
            assert('isset($limit[\'count\'])');
            $count = $limit['count'];
            assert('is_numeric($count)');

            // mysql version
            return " LIMIT $start, $count";
        }

        die('Wrong $limit parameter');
    }

    /**
     * Get the ORDER part of an SQL query
     */
    public function get_order($order) {
        if ($order == null) return '';
        return ' ORDER BY '.$order;
    }

    public function escape_string($str) {
        assert('false /* escape_string not implemented */');
    }

    public function last_insert_id($table = null) {
        assert('false /* last_insert_id not implemented */');
    }

    public function update($table, $record, $id) {
        $query = "UPDATE ".$this->quote_table_name($table)." SET ";
        $add_colon = false;
        foreach (array_keys($record) as $attr) {
            if ($add_colon) $query .= ', ';
            $query .= $this->quote_table_name($attr) .'=';
            if ($record[$attr] === null) {
                $query .= "NULL";
            } else if ($record[$attr] === true) {
                $query .= "TRUE";
            } else if ($record[$attr] === false) {
                $query .= "FALSE";
            } else {
                $query .= "'" . $this->escape_string($record[$attr]) . "'";
            }
            $query .= ' ';
            $add_colon = true;
        }
        $query .= "WHERE ".$this->quote_table_name('id')."=". $id . ";";

        // a least one change => query
        if (!$add_colon) return;

        // finally query
        $this->query($query);
    }

    public function insert($table, $record) {
        if (count($record)) {
            $query = 'INSERT INTO '.$this->quote_table_name($table).' (';
            $values = '';
            $add_colon = false;
            foreach (array_keys($record) as $attr) {
                if ($add_colon) {
                    $query .= ", ";
                    $values .= ", ";
                }
                $query .= $this->quote_table_name($attr) ;
                if ($record[$attr] === null) {
                    $values .= "NULL";
                } else if ($record[$attr] === true) {
                    $values .= "TRUE";
                } else if ($record[$attr] === false) {
                    $values .= "FALSE";
                } else if (is_int($record[$attr]) || is_float(($record[$attr]))) {
                    $values .= $this->escape_string($record[$attr]);
                } else {
                    $values .= "'". $this->escape_string($record[$attr])."'";
                }
                $add_colon = true;
            }
            $query .= ') VALUES (' . $values . ');';
        } else {
            // only default values
            $query = 'INSERT INTO '.$this->quote_table_name($table).' VALUES (DEFAULT);';
        }

        // finally query
        $this->query($query);

        // get this objects id
        return $this->last_insert_id($table);
    }

    /**
     * Get the WHERE part of an SQL query
     * @param $order = null means no order.
     *   $order = 'id DESC'
     */
    public function get_where($where) {
        if ($where == null) return '';
        if (is_string($where)) return ' WHERE '.$where;
        if (is_array($where)) {
            if (count($where) == 0) return '';
            $result = '';
            foreach ($where as $k => $v) {
                if (strlen($result) > 0) $result .= ' AND ';
                if (is_numeric($k)) {
                    // e.g. ...,'email_confirmed',...
                    $result .= $this->table_delimiter . mysql_escape_string($v)
                        . $this->table_delimiter;
                } else {
                    // e.g. ...,'email_confirmed' => 1,...
                    $result .= $this->table_delimiter . mysql_escape_string($k);
                    if (is_string($v)) {
                        $result .= $this->table_delimiter . ' = \''.$v.'\'';
                    } elseif (is_numeric($v)) {
                        $result .= $this->table_delimiter . ' = '.$v;
                    } elseif (is_bool($v)) {
                        $result .= $this->table_delimiter . ' = '.($v?'TRUE':'FALSE');
                    } elseif (is_null($v)) {
                        $result .= $this->table_delimiter . ' IS NULL';
                    } elseif ($v instanceof database_object) {
                        $result .= '_id'.$this->table_delimiter.' = '.($v->id);
                    } else {
                        die('sql WHERE type fail');
                    }
                }
            }
            return ' WHERE '.$result;
        }
        die('$where must be null, a string or an array');
    }

    /**
     * Those abstract interfaces have to be overwritten by sub-classes like 
     * db_adodb and db_mysql.
     */
    abstract public function query($query);
    abstract public function query_row($query);
    abstract public function fetch_columns($table);
    abstract public function select_single($query);
    abstract public function init();
}

