<?php
/**
 * AdoDB database adapter
 */
require_once('adodb/adodb.inc.php');

class db_adodb extends db_adapter {
    var $db = null;

    public function db_adodb($host_or_db, $user = null, $pass = null, $dbase = null) {
        $this->table_delimiter = '"';
        if ($host_or_db instanceof ADODB_postgres7) {
            // AdoDB object given => use this
            $this->db = $host_or_db;
            return;
        }

        // host / user / pass / db given
        assert('is_string($host_or_db)');
        assert('is_string($user)');
        assert('is_string($pass)');
        assert('is_string($dbase)');

        // connect to postgres7
        $this->db = NewADOConnection("postgres7://" . $user . ":" .
            $pass . "@" . $host_or_db . "/" . $dbase);
    }

    public function init() {
    }

    public function query($query) {   
        $this->log_query($query);
        $r = $this->db->Execute($query);
        if ($r === false) {
            // again, with debug
            $this->db->debug = true;
            $r = $this->db->Execute($query);
            assert('$r != false /* Query failed: ' . $query . ' */');
            die(); // in case that assert does not break
        }
        return $r;
    }

    public function query_row($query) {   
        $this->log_query($query);
        $r = $this->db->GetRow($query);
        if ($r === false) {
            // again, with debug
            $this->db->debug = true;
            $r = $this->db->GetAll($query);
            assert('$r != false /* Query failed: ' . $query . ' */');
            die(); // in case that assert does not break
        }
        foreach ($r as $k => $v) {
            if (is_numeric($k)) unset($r[$k]);
        }
        return $r;
    }

    /**
     *
     */
    public function get_limit($limit) {
        if (is_null($limit)) {
            return '';
        } else if (is_int($limit)) {
            return ' LIMIT '.$limit.' OFFSET 0';
        } else if (is_array($limit)) {
            if (isset($limit['start'])) {
                $start = $limit['start'];
            } else {
                $start = 0;
            }
            assert('is_numeric($start)');
            assert('isset($limit[\'count\'])');
            $count = $limit['count'];
            assert('is_numeric($count)');

            // mysql version
            return " LIMIT $count OFFSET $start";
        }

        die('Wrong $limit parameter');
    }

    public function query_all($query) {
        $this->log_query($query);
        $all = $this->db->GetAll($query);
        if ($all === false) {
            // again with debug
            $this->db->debug = true;
            $all = $this->db->GetAll($query);
            die("SQL query failed.");
        }
        // remove index values
        foreach ($all as $i => $o) {
            foreach ($o as $k => $v) {
                if (is_numeric($k)) unset($o[$k]);
            }
            $all[$i] = $o;
        }
        return $all;
    }

    private function auto_execute($table, $record, $action, $where = null) {
        // TODO AutoExecute does SELECT * before UPDATE => sux
        if (is_null($where)) {
            $r = $this->db->AutoExecute($table, $record, $action);
        } else {
            $r = $this->db->AutoExecute($table, $record, $action, $where); 
        }
        if ($r == false) {
            // again, with debug
            $this->db->debug = true;
            if (is_null($where)) {
                $r = $this->db->AutoExecute($table, $record, $action);
            } else {
                $r = $this->db->AutoExecute($table, $record, $action, $where); 
            }
            assert('$r != false /* query failed: AutoExecute '.$action.' on '.
                $table.' */');
        }
    }

    public function escape_string($str)
    {
        return pg_escape_string($str);
    }

    function quote_table_name($table_name) {
        return '"'.$this->escape_string($table_name).'"';
    }


    public function last_insert_id($table) {
        assert('is_string($table)');
        // geht nicht obwohl dokumentiert: (nachtrag: geht wohl nur mit OIDs)
        // $id = $this->db->Insert_ID();
        $query = 'SELECT id FROM '.$this->quote_table_name($table).' ORDER BY id DESC LIMIT 1';
        return $this->select_single($query);
    }


    /**
     * AdoDB implementation of insert uses AutoExecute
     * Removing this function would result in calling the db_adodb parent which 
     * would also work.
     * UPDATE: AdoDB fails if a column named e.g. 'select' has to be inserted
     * => so finally use own insert code
     */
    // public function insert($table, $record) {
    //     if (sizeof($record) == 0) {
    //         // AdoDB fails on insert with only default parameters
    //         $query = "INSERT INTO ".$table." VALUES (DEFAULT)";
    //         $r = $this->query($query);
    //     } else {
    //         // AdoDB fails on boolean parameters => use strings
    //         foreach ($record as $k => $v) {
    //             if ($v === false) $record[$k] = 'FALSE';
    //             else if ($v === true) $record[$k] = 'TRUE';
    //         }
    //         $this->auto_execute($table, $record, 'INSERT');
    //     }
    //     return $this->last_insert_id($table);
    // }

    /**
     * AdoDB implementation of update uses AutoExecute
     * Removing this function would result in calling the db_adodb parent which 
     * would also work.
     * UPDATE: AdoDB fails if a column named e.g. 'select' has to be updated
     * => so finally use own insert code
     */
    //public function update($table, $record, $id) {
    //    // AdoDB fails on boolean parameters => use strings
    //    foreach ($record as $k => $v) {
    //        if ($v === false) $record[$k] = 'FALSE';
    //        else if ($v === true) $record[$k] = 'TRUE';
    //    }
    //    $this->auto_execute($table, $record, 'UPDATE', 'id='.$id);
    //}

    /**
     * Optimize this table. Call e.g. from Admin interface or once on long 
     * program run.
     */
    public static function optimize($table) {
        // Not implemented for AdoDB. Just do nothing
    }

    public function fetch_columns($table) {
        $columns = array();
        $this->log_query('AdoDB::MetaColumns '.$table);
        $result = $this->db->MetaColumns($table);
        if ($result === false) die('MetaColumns failed on '.$table.'. Is table '.$table.' existent in db?');
        foreach ($result as $row) {
            $columns[$row->name] = $this->from_adodb_type($row->type);
        }
        return $columns;
    }

    /** 
     * convert postgresql type to php type
     */
    private function from_adodb_type($type) {
        switch ($type) {
        case 'int2':
        case 'int4':
            return 'integer';
        case 'bool': 
            return 'bool';
        case 'float4':
        case 'numeric':
            return 'float';
        case 'date':
            return 'date';
        case 'timestamp': 
            return 'datetime';
        case 'text':
        case 'bpchar':
        case 'varchar':
            return 'string';
        default: 
            echo('WARNING: which type is: '. $type.'? using string.');
            //die('Which type is: '. $type);
            return 'string';
        }
    }

    /**
     * Select one single number from the database (e.g. a count)
     */
    public function select_single($query) {
        $result = $this->db->GetAll($query);
        if ($result === false) {
            // again with debug
            $this->db->debug = true;
            $this->db->GetAll($query);
            die("SQL query failed.");
        }
        return $result[0][0];
    }
}

